%
%
%

Representing Colors as four-tuple of ints.  This rep avoids unnecessary
extra pack/unpack operations.

\begin{code}
module ColorByteTuple
	(
	 ColorE

	-- aToColor (a,r,g,b) = (a,a,a,a)
	, aToColor     -- :: ColorE -> ColorE

	, bgra         -- :: FloatE -> FloatE -> FloatE -> FloatE -> ColorE
        , colorToBGRA  -- :: ColorE -> (FloatE,FloatE,FloatE,FloatE)
	
	, fractToColor -- :: FractionE -> ColorE

        -- Int <-> Color conversions:
	, fromBGR32    -- :: BGR32E -> ColorE
	, fromBGR24    -- :: BGR24E -> ColorE
	, fromRGB24    -- :: IntE -> ColorE
        , fromBGRA32   -- :: BGRA32E -> ColorE

	, toBGR24       -- :: ColorE -> BGR24E
	, toRGB24       -- :: ColorE -> BGR24E
        , toBGRA32      -- :: ColorE -> BGRA32E
	, bgr32ToRGB24  -- :: BGR32E -> RGB24E
	, bgra32ToBGR32 -- :: BGR32E -> BGR24E
	, rgbaToBGRA    -- :: IntE -> ColorE

        , distC, distSqC  -- :: ColorE -> ColorE -> FloatE

	  -- basic color combinators:
	, negC	       -- :: ColorE -> ColorE
	, lerpC        -- :: FractionE -> ColorE -> ColorE -> ColorE
	, fadeC	       -- :: FractionE -> ColorE -> ColorE
	, overC        -- :: ColorE -> ColorE -> FloatE

	, usesSSE 

        , colorVar     -- :: String -> ColorE
	
	, isPreMultiplied
	) where

import Exp
import Bits
import List (intersperse)
import IOExts (trace)

\end{code}


The color representation is here a BGRA tuple of bytes.

\begin{code}

-- Four-tuple of byte values.  Order: B,G,R,A, as in memory order.  
type ColorE = (ByteFracE, ByteFracE, ByteFracE, ByteFracE)

-- To signal to the backend whether we're using SSE (=> use the Intel C
-- compiler), all implementations of Color export the following flag:
usesSSE :: Bool
usesSSE = False
\end{code}


\begin{code}

bgra :: FractionE -> FractionE -> FractionE -> FractionE -> ColorE
bgra b g r a = (toBF b, toBF g, toBF r, toBF a)

colorToBGRA :: ColorE -> (FloatE,FloatE,FloatE,FloatE)
colorToBGRA (b,g,r,a) = (fromBF b,fromBF g,fromBF r,fromBF a)


fractToColor :: FractionE -> ColorE
fractToColor x = let bf = toBF x in (bf,bf,bf,bf)
-- Equivalent, but would recompute toBF x:
-- fractToColor x = argb x x x x

aToColor :: ColorE -> ColorE
aToColor (b,g,r,a) = (a,a,a,a)


-- Conversions

fromBGRA32 :: BGR32E -> ColorE
fromBGR32  :: BGR32E -> ColorE
fromBGR24  :: BGR24E -> ColorE
fromRGB24  :: RGB24E -> ColorE

fromBGRA32 bgra32 = 
--  ifE (a >=* b &&* a >=* g &&* a >=* r)
      (b,g,r,a)
--      (a,a,a,a)
  where
   (b,g,r,a) = unpack4 bgra32  

isPreMultiplied :: ColorE -> BoolE
isPreMultiplied (b,g,r,a) = a >=* b &&* a >=* g &&* a >=* r

fromBGR32 bgr32 = (b,g,r,255) where (b,g,r,_) = unpack4 bgr32
fromBGR24 bgr24 = (b,g,r,255) where (b,g,r,_) = unpack4 bgr24
fromRGB24 rgb24 = (b,g,r,255) where (r,g,b,_) = unpack4 rgb24

toBGRA32 :: ColorE -> BGRA32E
-- The "reserved" byte must be 0, not 255.
toBGR24 :: ColorE -> BGR24E
toRGB24 :: ColorE -> RGB24E

bgr32ToRGB24 bgr32 = packL [r,g,b] where (b,g,r,_) = unpack4 bgr32

bgra32ToBGR32 bgr32 = packL [b,g,r,0] where (b,g,r,_) = unpack4 bgr32
rgbaToBGRA rgb32 = (b,g,r,a) where (r,g,b,a) = unpack4 rgb32


toBGRA32 (b,g,r,a) = packL [b,g,r,a]
toBGR24  (b,g,r,_) = packL [b,g,r]
toRGB24  (b,g,r,_) = packL [r,g,b]

\end{code}

Color combinators that are 'better' (aka. more efficiently) expressed
in terms of the concrete Color representation.

\begin{code}
-- Dot product for colors
dotC :: ColorE -> ColorE -> FloatE
dotC (b1,g1,r1,a1) (b2,g2,r2,a2) =
  fromBF a1 * fromBF a2 + 
  fromBF r1 * fromBF r2 + 
  fromBF g1 * fromBF g2 + 
  fromBF b1 * fromBF b2


-- Distance and distance squared in ARGB space
distC, distSqC :: ColorE -> ColorE -> FloatE

distSqC (b1,g1,r1,a1) (b2,g2,r2,a2) =
  dotC d d where d = (b1-b2,g1-g2,r1-r2,a1-a2)

distC c c' = sqrt (distSqC c c')


-- Color negative.
negC :: ColorE -> ColorE
negC (b,g,r,a) = (a-b,a-g,a-r,a)

lerpC :: FractionE -> ColorE -> ColorE -> ColorE
lerpC w (b1,g1,r1,a1) (b2,g2,r2,a2) =
  (h b1 b2, h g1 g2, h r1 r2, h a1 a2)
 where
   h = lerp (toBF w)
   lerp c a b = --trace (showCall "lerp" [c,a,b] ++ "\n") $
                a + multBF c (b - a)

showCall fname args = concat $ intersperse " " $ fname : map showP args
 where
   showP x = "(" ++ show x ++ ")"

fadeC :: FractionE -> ColorE -> ColorE
fadeC f (b,g,r,a) = (h b, h g, h r, h a)
 where
   h = multBF (toBF f)

overC :: ColorE -> ColorE -> ColorE
overC (b1,g1,r1,a1) (b2,g2,r2,a2) =
  (h b1 b2, h g1 g2, h r1 r2, h a1 a2)
 where
   h = ablend a1

   -- Alpha blending for pre-multiplied values.  See tech memo #4, from
   -- http://www.research.microsoft.com/~alvy/Memos/default.htm.
   ablend aTop top bot = top + multBF (255 - aTop) bot


-- For testing

-- Colors are not expressions, so make four variables

colorVar :: String -> ColorE
colorVar name = (v 'b', v 'g', v 'r', v 'a')
 where
   v ch = var (ch : name)


\end{code}
